/*
 * DECthreads example program featuring an asynchronous user interface
 */
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <pthread.h>
#include <exc_handling.h>
#include <utils.h>

/*
 * Local definitions
 */
#define PROMPT     "Info> "
#define MAXLINSIZ  81
#define THDNUM     5

/*
 * Server thread "states"
 */
#define ST_INIT        0
#define ST_FINISHED    1
#define ST_CANCELED    2
#define ST_ERROR       3
#define ST_RUNNING     4

/*
 * Global variables
 */
struct THREAD_DATA
{
    pthread_t       thread_handle;
    pthread_mutex_t mutex_handle;
    int             time;
    char            task;
    int             state;
};

static struct THREAD_DATA thread_data[THDNUM];
static pthread_mutex_t free_thread_mutex;
static pthread_cond_t free_thread_cv;
static int      free_thread;

/*
 * Local Routines
 */
static void
dispatch_task(pthread_startroutine_t routine, char task, int time);

static void
do_cancel(int index);

static void
do_cleanup(int index, int final_state);

static pthread_addr_t
do_date(pthread_addr_t arg);

static void
do_delay(int index);

static void
do_status(int index);

static pthread_addr_t
do_time(pthread_addr_t arg);

static void
do_wait(int index);

static int
find_free_thread(int *index);

static char *
get_cmd(char *buffer);

static int
get_y_or_n(char *query, char defans);

static void
init_routine(void);

static void
print_error(char *message);

static void
print_help(void);

static void
this_shouldnt_happen(char *message);

static pthread_once_t once_block = pthread_once_init;

/*
 * The main program:  
 */
int
main(int argc, char *argv[])
{
    int             done = FALSE;
    char            cmdline[MAXLINSIZ];
    char            cmd_wd[MAXLINSIZ];
    int             cmd_arg;
    int             cmd_cnt;
    pthread_startroutine_t routine;

    /*
     * Perform program initialization.
     */
    pthread_once(&once_block, init_routine);

    /*
     * Main command loop
     */
    do
    {
	/*
	 * Get and parse a command.  Yield first so that any threads waiting
	 * to execute get a chance to before we take out the global lock 
	 * and block for I/O.
	 */
	pthread_yield();
	if (get_cmd(cmdline))
	{
	    cmd_cnt = sscanf(cmdline, "%s %d", cmd_wd, &cmd_arg);
	    routine = (pthread_startroutine_t) NULL;

	    if ((cmd_cnt == 1) || (cmd_cnt == 2))
	    {
		cmd_wd[0] = tolower(cmd_wd[0]);
		switch (cmd_wd[0])
		{
		    case 'h':
		    case '?':
			print_help();
			break;

		    case 'q':
			done = TRUE;
			break;

		    case 's':
			do_status((cmd_cnt == 2 ? cmd_arg : -1));
			break;

		    case 'c':
		    case 'd':
		    case 't':
		    case 'w':
			if (cmd_cnt != 2)
			    print_error("Missing command argument.");
			else
			{
			    switch (cmd_wd[0])
			    {
				case 'c':
				    do_cancel(cmd_arg);
				    break;
				case 'd':
				    routine = do_date;
				    break;
				case 't':
				    routine = do_time;
				    break;
				case 'w':
				    do_wait(cmd_arg);
				    break;
				default:
				    this_shouldnt_happen("Inner switch.");
				    break;
			    }
			}
			break;
		    default:
			print_error("Unrecognized command.");
			break;
		}
	    }
	    else if (cmd_cnt != EOF)
		this_shouldnt_happen("Unexpected parse error.");


	    /*
	     * If there is a routine to be executed in a server thread,
	     * create the thread.
	     */
	    if (routine)
		dispatch_task(routine, cmd_wd[0], cmd_arg);
	}
	else
	    done = TRUE;
    }
    while (!done);

    return (EXIT_SUCCESS);
}


/*
 * Create a thread to handle the user's request.
 */
static void
dispatch_task(pthread_startroutine_t routine, char task, int time)
{
    int             i;

    if (find_free_thread(&i))
    {
	/*
	 * Record the data for this thread where both the main thread and the
	 * server thread can share it.  Lock the mutex to ensure exclusive
	 * access to the storage.
	 */
	pthread_mutex_lock(&thread_data[i].mutex_handle);
	thread_data[i].time = time;
	thread_data[i].task = task;
	thread_data[i].state = ST_RUNNING;
	pthread_mutex_unlock(&thread_data[i].mutex_handle);

	/*
	 * Create the thread, using the default attributes.  The thread will
	 * execute the specified routine and get its data from array slot 'i'.
	 */
	pthread_create(&thread_data[i].thread_handle,
		       pthread_attr_default,
		       routine,
		       (pthread_addr_t) i);

	printf_r("This is command #%d.\n\n", i);
    }
}

/*
 * Wait for the completion of the specified command.
 */
static void
do_cancel(int index)
{
    int             cancelable;

    if ((index < 0) || (index >= THDNUM))
	print_error("Bad command number.");
    else
    {
	pthread_mutex_lock(&thread_data[index].mutex_handle);
	cancelable = (thread_data[index].state == ST_RUNNING);
	pthread_mutex_unlock(&thread_data[index].mutex_handle);

	if (cancelable)
	    pthread_cancel(thread_data[index].thread_handle);
	else
	    print_error("Command is not active.");
    }
}

/*
 * Post-task clean-up routine.
 */
static void
do_cleanup(int index, int final_state)
{
    /*
     * This thread is about to make the change from "running" to "finished", 
     * so lock a mutex to prevent a race condition in which the main thread 
     * sees this thread as finished before it is actually done cleaning up.
     *
     * Note that when attempting to lock more than one mutex at a time, 
     * always lock the mutexes in the same order everywhere in the code.  
     * The ordering here is the same as in "find_free_thread".
     */
    pthread_mutex_lock(&free_thread_mutex);

    /*
     * Mark the thread as finished with its task.
     */
    pthread_mutex_lock(&thread_data[index].mutex_handle);
    thread_data[index].state = final_state;
    pthread_mutex_unlock(&thread_data[index].mutex_handle);

    /*
     * Set the flag indicating that there is a free thread, and signal the
     * main thread, in case it is waiting.
     */
    free_thread = TRUE;
    pthread_cond_signal(&free_thread_cv);
    pthread_mutex_unlock(&free_thread_mutex);
}

/*
 * Thread routine that prints out the date.
 *
 * Synchronize access to ctime as it is not thread-safe (it returns the address
 * of a static string).  Also synchronize access to stdio routines.
 */
static          pthread_addr_t
do_date(pthread_addr_t arg)
{
    time_t          clock_time;
    char           *date_str;
    char            day[4], month[4], date[3], year[5];
    int             index = (int) arg;

    TRY
    {
	/*
	 * Pretend that this task actually takes a long time to perform.
	 */
	do_delay(index);

	pthread_lock_global_np();
	TRY
	{
	    clock_time = time((time_t *) 0);
	    date_str = ctime(&clock_time);
	    sscanf(date_str, "%s %s %s %*s %s", day, month, date, year);
	    printf_r("[%d] Today is %s, %s %s %s.\n\n", index, day, date, month, year);
	}
	FINALLY
	{
	    pthread_unlock_global_np();
	}
	ENDTRY
    }

    CATCH(pthread_cancel_e)
    {
	printf_r("[%d] Canceled.\n", index);

	/*
	 * Perform exit actions
	 */
	do_cleanup(index, ST_CANCELED);

	RERAISE;
    }

    CATCH_ALL
    {
	printf_r("[%d] ", index);
	exc_report(THIS_CATCH);

	/*
	 * Perform exit actions
	 */
	do_cleanup(index, ST_ERROR);

	RERAISE;
    }
    ENDTRY

    /*
     * Perform exit actions (thread was not canceled).
     */
    do_cleanup(index, ST_FINISHED);

    /*
     * All thread routines return a value.  This program doesn't check the
     * value, however.
     */
    return ((pthread_addr_t) index);
}

/*
 * Delay routine
 *
 * Since the actual tasks that threads do in this program take so little time
 * to perform, execute a delay to make it seem like they are taking a long 
 * time.  Also, this will give the user something to query the progress of.
 */
static void
do_delay(int index)
{
    static struct timespec interval =
    {1, 0};
    int             done;

    while (TRUE)
    {
	/*
	 * Decrement the global count, so the main thread can see how much
	 * progress we've made.  Keep decrementing as long as the remaining
	 * time is greater than zero.
	 *
	 * Lock the mutex to ensure no conflict with the main thread that
	 * might be reading the time remaining while we're decrementing it.
	 */
	pthread_mutex_lock(&thread_data[index].mutex_handle);
	done = ((thread_data[index].time--) <= 0);
	pthread_mutex_unlock(&thread_data[index].mutex_handle);

	/*
	 * Quit if the time is up.
	 */
	if (done)
	    break;

	/*
	 * Wait for one second.
	 */
	pthread_delay_np(&interval);
    }
}

/*
 * Print the status of the specified thread.
 */
static void
do_status(int index)
{
    int             start, end;
    int             i;
    int             output = FALSE;

    if ((index < -1) || (index >= THDNUM))
	print_error("Bad command number.");
    else
    {
	if (index == -1)
	    start = 0, end = THDNUM;
	else
	    start = index, end = start + 1;

	for (i = start; i < end; i++)
	{
	    pthread_mutex_lock(&thread_data[i].mutex_handle);
	    if (thread_data[i].state != ST_INIT)
	    {
		printf_r("Command #%d:  ", i);

		switch (thread_data[i].task)
		{
		    case 't':
			printf_r("\"time\", ");
			break;
		    case 'd':
			printf_r("\"date\", ");
			break;
		    default:
			printf_r("[unknown] ");
			break;
		}

		switch (thread_data[i].state)
		{
		    case ST_FINISHED:
			printf_r("completed");
			break;
		    case ST_CANCELED:
			printf_r("canceled");
			break;
		    case ST_ERROR:
			printf_r("terminated by error");
			break;
		    case ST_RUNNING:
			if (thread_data[i].time < 0)
			    printf_r("waiting to print");
			else
			    printf_r("%d seconds remaining", thread_data[i].time);
			break;
		    default:
			this_shouldnt_happen("Thread state default");
			break;
		}

		printf_r(".\n");
		output = TRUE;
	    }

	    pthread_mutex_unlock(&thread_data[i].mutex_handle);
	}

	if (!output)
	    printf_r("No such command.\n");

	printf_r("\n");
    }
}


/*
 * Thread routine that prints out the date.
 */
static          pthread_addr_t
do_time(pthread_addr_t arg)
{
    time_t          clock_time;
    char           *date_str;
    char            time_str[8];

    TRY
    {
	/*
	 * Pretend that this task actually takes a long time to perform.
	 */
	do_delay((int) arg);
	pthread_lock_global_np();

	TRY
	{
	    clock_time = time((time_t *) 0);
	    date_str = ctime(&clock_time);
	    sscanf(date_str, "%*s %*s %*s %s", time_str);
	    printf_r("[%p] At the tone the time will be, %s.%c\n\n", arg, time_str, '\007');
	}
	FINALLY
	{
	    pthread_unlock_global_np();
	}
	ENDTRY
    }
    CATCH(pthread_cancel_e)
    {
	printf_r("[%p] Canceled.\n", arg);
	do_cleanup((int) arg, ST_CANCELED);
	RERAISE;
    }

    CATCH_ALL
    {
	printf_r("[%p] ", arg);
	exc_report(THIS_CATCH);
	do_cleanup((int) arg, ST_ERROR);
	RERAISE;
    }
    ENDTRY

    /*
     * Perform exit actions (thread was not canceled).
     */
    do_cleanup((int) arg, ST_FINISHED);

    /*
     * All thread routines return a value.  This program doesn't check the
     * value, however.
     */
    return (arg);
}


/*
 * Wait for the completion of the specified command.
 */
static void
do_wait(int index)
{
    if ((index < 0) || (index >= THDNUM))
	print_error("Bad command number.");
    else
	pthread_join(thread_data[index].thread_handle, (pthread_addr_t) NULL);
}

/*
 * Find a free server thread to handle the user's request.
 *
 * If a free thread is found, its index is written at the supplied address
 * and the function returns true.
 */
static int
find_free_thread(int *index)
{
    int             i;
    int             found;
    int             retry = FALSE;

    do
    {
	/*
	 * We're about to look for a free thread, so prevent the data state 
	 * from changing while we are looking.
	 *
	 * Note that when attempting to lock more than one mutex at a time, 
	 * always lock the mutexes in the same order everywhere in the code.  
	 * The ordering here is the same as in "do_cleanup".
	 */
	pthread_mutex_lock(&free_thread_mutex);

	/*
	 * Find a slot that doesn't have a running thread in it.
	 *
	 * Before checking, lock the mutex to prevent conflict with the thread
	 * if it is running.
	 */
	for (i = 0, found = FALSE; i < THDNUM; i++)
	{
	    pthread_mutex_lock(&thread_data[i].mutex_handle);
	    found = (thread_data[i].state != ST_RUNNING);
	    pthread_mutex_unlock(&thread_data[i].mutex_handle);


	    /*
	     * Now that the mutex is unlocked, break out of the loop if the
	     * thread is free.
	     */
	    if (found)
		break;
	}

	if (found)
	    retry = FALSE;
	else
	{
	    retry = get_y_or_n("All threads are currently busy, do you want to wait?", 'Y');

	    if (retry)
	    {
		/*
		 * All threads were busy when we started looking, so clear
		 * the "free thread" flag.
		 */
		free_thread = FALSE;

		/*
		 * Now wait until some thread finishes and sets the flag
		 */
		while (!free_thread)
		    pthread_cond_wait(&free_thread_cv, &free_thread_mutex);
	    }
	}

	pthread_mutex_unlock(&free_thread_mutex);
    }
    while (retry);


    if (found)
    {
	/*
	 * Request DECthreads reclaim its internal storage for this old thread
	 * before we use the handle to create a new one.
	 */
	pthread_detach(&thread_data[i].thread_handle);

	*index = i;
    }

    return (found);
}

/*
 * Get the next user command.
 *
 * Synchronize I/O with other threads to prevent conflicts if the stdio 
 * routines are not thread-safe.
 */
static char    *
get_cmd(char *buffer)
{
    char           *ret_val;

    /*
     * The following calls to stdio routines are protected from concurrent 
     * calls by other threads by locking the "global lock".
     */
    printf_r(PROMPT);
    ret_val = gets(buffer);
    return ret_val;
}

/*
 * Get a yes or no answer to a query.  A "blank" answer uses default answer.
 *
 * Synchronize I/O with other threads to prevent conflicts if the stdio 
 * routines are not thread-safe.
 *
 * Returns TRUE for "yes" and FALSE for "no".
 */
static int
get_y_or_n(char *query, char defans)
{
    char            buffer[MAXLINSIZ];
    int             answer;
    int             retry = TRUE;

    do
    {
	buffer[0] = '\0';
	printf_r("%s [%c] ", query, defans);
	gets(buffer);

	if (buffer[0] == '\0')
	    buffer[0] = defans;

	switch (buffer[0])
	{
	    case 'y':
	    case 'Y':
		answer = TRUE;
		retry = FALSE;
		break;
	    case 'n':
	    case 'N':
		answer = FALSE;
		retry = FALSE;
		break;
	    default:
		printf_r("Please enter \"Y\" or \"N\".\n");
		retry = TRUE;
		break;
	}
    }
    while (retry);

    return answer;
}


/*
 * Initialization routine;
 *
 * Called as a one-time initialization action.
 */
static void
init_routine(void)
{
    int             i;

    for (i = 0; i < THDNUM; i++)
    {
	pthread_mutex_init(&thread_data[i].mutex_handle, pthread_mutexattr_default);
	thread_data[i].time = 0;
	thread_data[i].task = '\0';
	thread_data[i].state = ST_INIT;
    }

    pthread_mutex_init(&free_thread_mutex, pthread_mutexattr_default);
    pthread_cond_init(&free_thread_cv, pthread_condattr_default);
}

/*
 * Print user-error message.
 *
 * Synchronize I/O with other threads to prevent conflicts if the stdio 
 * routines are not thread-safe.
 */
static void
print_error(char *message)
{
    printf_r("%s\n\n", message);
}



/*
 * Print help text.
 *
 * Synchronize I/O with other threads to prevent conflicts if the stdio 
 * routines are not thread-safe.
 */

static void
print_help(void)
{
    printf_r("Commands are formed by a verb and an optional numeric argument.\n");
    printf_r("The following commands are available:\n");
    printf_r("\tCancel\t[command]\tCancel running command\n");
    printf_r("\tDate\t[delay]\t\tPrint the date\n");
    printf_r("\tHelp\t\t\tPrint this text\n");
    printf_r("\tQuit\t\t\tQuit (same as EOF)\n");
    printf_r("\tStatus\t[command]\tReport on running command\n");
    printf_r("\tTime\t[delay]\t\tPrint the time\n");
    printf_r("\tWait\t[command]\tWait for command to finish\n");
    printf_r("\n[command] refers to the command number.\n");;
    printf_r("[delay] delays the command execution for some number of seconds.\n");
    printf_r("This delay simulates a command task that actually takes some\n");
    printf_r("period of time to execute.  During this delay, commands may be\n");
    printf_r("initiated, queried, and/or canceled.\n");
}

/*
 * Print error message for unexpected execution event.
 *
 * Synchronize I/O with other threads to prevent conflicts if the stdio 
 * routines are not thread-safe.
 */
static void
this_shouldnt_happen(char *message)
{
    printf_r("\n\nThis shouldn't happen: %s\n\n", message);
}
